// Copyright (c) 2013-2019 by Michael Dvorkin and contributors. All Rights Reserved.
// Use of this source code is governed by a MIT-style license that can
// be found in the LICENSE file.

package main

import (
	"flag"
	"fmt"
	"os"
	"os/user"
	"path"
	"strings"
	"time"
	"net/http"
	"bufio"
	"encoding/json"
	"strconv"
	//"log"
	//"io/ioutil"

	"easyquotation"
	"easyquotation/sina"
	"easyquotation/stock"
	"github.com/gocolly/colly"

	"github.com/eiannone/keyboard"
	"github.com/mop-tracker/mop"
	"github.com/nsf/termbox-go"
	"github.com/olekukonko/tablewriter"

	mqtt "github.com/eclipse/paho.mqtt.golang"
)

// File name in user's home directory where we store the settings.
const defaultProfile = `.moprc`

const help = `Mop v1.0.0 -- Copyright (c) 2013-2022 by Michael Dvorkin. All Rights Reserved.
NO WARRANTIES OF ANY KIND WHATSOEVER. SEE THE LICENSE FILE FOR DETAILS.

<u>Command</u>    <u>Description                                </u>
   +       Add stocks to the list.
   -       Remove stocks from the list.
   ?       Display this help screen.
   f       Set filtering expression.
   F       Unset filtering expression.
   g       Group stocks by advancing/declining issues.
   o       Change column sort order.
   p       Pause market data and stock updates.
   Scroll  Scroll up/down.
   PgUp/PgDn; Up/Down arrow; j/k;J/K also all scroll up/down
   q       Quit mop.
  esc      Ditto.

Enter comma-delimited list of stock tickers when prompted.

<r> Press any key to continue </r>
`


func getuserinput(preset *Preset, sc mop.Stock) string{
	scanner := bufio.NewScanner(os.Stdin)
	var precondition string
	var instrumentname string
	var buyorsell string

	if sc.Ticker == "" {
		fmt.Print("Enter Command: ")
		scanner.Scan()
		cmdstr := scanner.Text()

		inputcmd := strings.Split(cmdstr, " ")
		if len(inputcmd) == 2 {
			if inputcmd[0] == "buy" {
				preset.Direction = 23
			}else if inputcmd[0] == "sell" {
				preset.Direction = 24
			}else{
				return ""
			}
		}else{
			return ""
		}

		preset.Scode = inputcmd[1]
	}else{
		preset.Direction = 23
		preset.Scode = sc.Dividend[2:]
		f1, _ := strconv.ParseFloat(sc.LastTrade, 64)
		f2, _ := strconv.ParseFloat(sc.Change, 64)
		precondition = fmt.Sprintf("%s>%.2f", sc.LastTrade, f1-f2)
		instrumentname = sc.Ticker[2:]
	}

	if preset.Direction == 23 {
		buyorsell = "buy"
	} else {
		buyorsell = "sell"
	}

	if preset.Ifbelow == 0 && preset.Ifabove == 0  {

	}else if preset.Ifbelow == 0 {
		precondition = fmt.Sprintf(">%f", preset.Ifabove)
	}else{
		precondition = fmt.Sprintf("<%f", preset.Ifbelow)
	}

	fmt.Printf("Enter Condition[%s %s][%s]: ", buyorsell, preset.Scode+instrumentname, precondition)

	scanner.Scan()
	condition := scanner.Text()
	if strings.Contains(condition, ">") {
		preset.Ifbelow = 0
		preset.Ifabove, _ = strconv.ParseFloat(strings.Replace(condition, ">", "", -1), 64)
	}else if strings.Contains(condition, "<") {
		preset.Ifabove = 0
		preset.Ifbelow, _ = strconv.ParseFloat(strings.Replace(condition, "<", "", -1), 64)
	}else{
		return ""
	}

	if preset.Ifbelow == 0 && preset.Ifabove == 0  {
		precondition = ""
	}else if preset.Ifbelow == 0 {
		precondition = fmt.Sprintf(">%.2f", preset.Ifabove)
	}else{
		precondition = fmt.Sprintf("<%.2f", preset.Ifbelow)
	}
	fmt.Printf("Enter Vol[%s %s][%s]: ", buyorsell, preset.Scode+instrumentname, precondition)
	scanner.Scan()
	vol := scanner.Text()
	preset.Vol, _ = strconv.ParseFloat(vol, 64)
	if preset.Vol == 0 {
		return ""
	}

	//fmt.Print(preset)
	jsonData, err := json.Marshal(preset)
	if err != nil {
		fmt.Println(err)
	}
	extraField := `,"cmd": "preset"}`
	jsonData = append(jsonData[:len(jsonData)-1], extraField...)
	message := string(jsonData)
	fmt.Println(message)

	return message
}

//-----------------------------------------------------------------------------
func mainLoop(screen *mop.Screen, profile *mop.Profile, mode string) {
	var lineEditor *mop.LineEditor
	var columnEditor *mop.ColumnEditor

	termbox.SetInputMode(termbox.InputMouse)

	// use buffered channel for keyboard event queue
	keyboardQueue := make(chan termbox.Event, 128)

	timestampQueue := time.NewTicker(1 * time.Second)
	quotesQueue := time.NewTicker(2 * time.Second)
	marketQueue := time.NewTicker(12 * time.Second)
	showingHelp := false
	paused := false
	upDownJump := profile.UpDownJump
	redrawQuotesFlag := false
	redrawMarketFlag := false
	account := ""
	var canstock mop.Stock

	// 创建一个用于存储用户输入的缓冲区
    input := ""
	easyquotation.Init()
	c := colly.NewCollector()
	SinaStock_spider := sina.NewSinaStock(c)
	si := stock.G_STOCK_MANAGER.StockList
	if mode == "review" { 
		SinaStock_spider.Start("")
	}else{
		SinaStock_spider.Start("stock_in.json")
	}

	go func() {
		for {
			keyboardQueue <- termbox.PollEvent()
		}
	}()

	// create a new MQTT client
	opts := mqtt.NewClientOptions()
	opts.AddBroker("tcp://119.29.166.226:1883")
	currentTime := time.Now().UnixNano() / int64(time.Millisecond)
    millisecondsString := strconv.FormatInt(currentTime, 10)
    opts.SetClientID("client-mop-"+millisecondsString)
    opts.SetUsername("sandy")
    opts.SetPassword("Abb123456.")

	client := mqtt.NewClient(opts)

	// connect to the MQTT broker
	if token := client.Connect(); token.Wait() && token.Error() != nil {
		panic(token.Error())
	}

	// create a channel for receiving MQTT messages
	messages := make(chan mqtt.Message)

	// subscribe to the MQTT topic of interest
	//topic := "stock/response"
	topic := "stock/response/#"
	if token := client.Subscribe(topic, 0, func(client mqtt.Client, message mqtt.Message) {
		//fmt.Printf("Received message: %s from topic: %s\n", message.Payload(), message.Topic())
		// send the message to the messages channel
		messages <- message
	}); token.Wait() && token.Error() != nil {
		panic(token.Error())
	}

	response, err := http.Get("http://119.29.166.226/q/dayjson/ml.json")//("http://119.29.166.226/q/dayjson/ml.json")
	if err != nil {
		// Handle error
		fmt.Println(err)
	}
	defer response.Body.Close()
	var watchlist mop.Watchlist
	err = json.NewDecoder(response.Body).Decode(&watchlist)
	if err != nil {
		fmt.Println(err)
	}
	market := mop.NewMarket(stock.G_STOCK_MANAGER.StockList, &watchlist)
	
	quotes := mop.NewQuotes(market, profile, stock.G_STOCK_MANAGER.StockList, &watchlist, client)
	codestoadd := []string{"sh000001", "sz399001", "sz399006"}
	quotes.Addstockcodetofile(codestoadd)
	screen.Draw(market)
	screen.Draw(quotes)

loop:
	for {
		select {
		case event := <-keyboardQueue:
			switch event.Type {
			case termbox.EventKey:
				if lineEditor == nil && columnEditor == nil && !showingHelp {
					if event.Key == termbox.KeyEsc || event.Ch == 'q' || event.Ch == 'Q' {
						// 取消订阅并断开连接
						if token := client.Unsubscribe(topic); token.Wait() && token.Error() != nil {
							panic(token.Error())
						}
						client.Disconnect(250)
						quotes.SaveStocks()
						break loop
					} else if event.Ch == '+' || event.Ch == '-' {
						lineEditor = mop.NewLineEditor(screen, quotes)
						lineEditor.Prompt(event.Ch)
					} else if event.Ch == 'f' {
						lineEditor = mop.NewLineEditor(screen, quotes)
						lineEditor.Prompt(event.Ch)
					} else if event.Ch == 'F' {
						profile.SetFilter("")
					} else if event.Ch == 'o' || event.Ch == 'O' {
						columnEditor = mop.NewColumnEditor(screen, quotes)
					} else if event.Ch == 'g' || event.Ch == 'G' {
						if profile.Regroup() == nil {
							screen.Draw(quotes)
						}
					} else if event.Ch == 'p' || event.Ch == 'P' {
						paused = !paused
						screen.Pause(paused).Draw(time.Now())
					} else if event.Ch == '?' || event.Ch == 'h' || event.Ch == 'H' {
						showingHelp = true
						screen.Clear().Draw(help)
					} else if event.Key == termbox.KeyPgdn ||
						event.Ch == 'J' {
						screen.IncreaseOffset(upDownJump)
						redrawQuotesFlag = true
					} else if event.Key == termbox.KeyPgup ||
						event.Ch == 'K' {
						screen.DecreaseOffset(upDownJump)
						redrawQuotesFlag = true
					} else if event.Key == termbox.KeyArrowUp || event.Ch == 'k' {
						screen.DecreaseOffset(1)
						screen.Selectmoveup(quotes)
						selstock := quotes.Getselectedinfo(screen.Selectindex())
						if selstock != nil {
							canstock = *selstock
						}else{
							canstock = mop.Stock{}
						}
						redrawQuotesFlag = true
					} else if event.Key == termbox.KeyArrowDown || event.Ch == 'j' {
						screen.IncreaseOffset(1)
						screen.Selectmovedown(quotes)
						selstock := quotes.Getselectedinfo(screen.Selectindex())
						if selstock != nil {
							canstock = *selstock
						}else{
							canstock = mop.Stock{}
						}
						redrawQuotesFlag = true
					} else if event.Key == termbox.KeyHome {
						screen.ScrollTop()
						redrawQuotesFlag = true
					} else if event.Key == termbox.KeyEnd {
						screen.ScrollBottom()
						redrawQuotesFlag = true
					} else if event.Ch == 'b' {
						data := map[string]string{
							"cmd": "getaccount",
						}
						jsonData, err := json.Marshal(data)
						if err != nil {
							fmt.Println(err)
						}
						message := string(jsonData)
						token := client.Publish("stock/request/470100037961", 0, false, message)
						token.Wait()
						account = "470100037961"
					} else if event.Key == termbox.KeySpace {
						data := map[string]string{
							"cmd": "getaccount",
						}
						jsonData, err := json.Marshal(data)
						if err != nil {
							fmt.Println(err)
						}
						message := string(jsonData)
						token := client.Publish("stock/request/620000301588", 0, false, message)
						token.Wait()
						account = "620000301588"
					} else if event.Ch >= '0' && event.Ch <= '9' {
						input += string(event.Ch)
					} else if event.Key == termbox.KeyEnter || event.Key == termbox.KeyArrowLeft || event.Key == termbox.KeyArrowRight {
						indexnum, err := strconv.Atoi(input)
						selectindex := screen.Selectindex()
						if selectindex > 0 {
							selcode := canstock.Dividend
							if event.Key == termbox.KeyArrowLeft || event.Key == termbox.KeyEnter {
								quotes.Sendstockgraphreq(selcode, false)
							}
							if event.Key == termbox.KeyArrowRight || event.Key == termbox.KeyEnter {
								quotes.Sendstockgraphreq(selcode, true)
							}
						}else{
						//fmt.Println("indexnum:", indexnum)
							if err == nil && indexnum > 0 {
								if event.Key == termbox.KeyArrowLeft || event.Key == termbox.KeyEnter {
									quotes.Sendstockgraphreq(indexnum, false)
								}
								if event.Key == termbox.KeyArrowRight || event.Key == termbox.KeyEnter {
									quotes.Sendstockgraphreq(indexnum, true)
								}
							}
						}
						// 清空输入缓冲区
						input = ""
					}
				} else if lineEditor != nil {
					if done := lineEditor.Handle(event); done {
						lineEditor = nil
					}
				} else if columnEditor != nil {
					if done := columnEditor.Handle(event); done {
						columnEditor = nil
					}
				} else if showingHelp {
					showingHelp = false
					screen.Clear().Draw(market, quotes)
				}
			case termbox.EventResize:
				screen.Resize()
				if !showingHelp {
					//screen.Draw(market)
					//redrawQuotesFlag = true
					//screen.Draw(market)
					redrawQuotesFlag = true
					redrawMarketFlag = true
					//screen.DrawOldQuotes(quotes)
				} else {
					screen.Draw(help)
				}
			case termbox.EventMouse:
				if lineEditor == nil && columnEditor == nil && !showingHelp {
					switch event.Key {
					case termbox.MouseWheelUp:
						screen.DecreaseOffset(5)
						redrawQuotesFlag = true
					case termbox.MouseWheelDown:
						screen.IncreaseOffset(5)
						redrawQuotesFlag = true
					}
				}
			}

		case <-timestampQueue.C:
			now := time.Now()
			//add for save stock records
			hour := now.Hour()
			minute := now.Minute()
			second := now.Second()

			if hour > 9 && hour < 16 {
				// Check if the current time is 1 minute and 0 seconds past every hour
				if minute == 1 && second == 0 {
					//fmt.Println("It's 1 minute and 0 seconds past the hour")
					quotes.SaveStocks()
				} 
			} 

			if !showingHelp && !paused {
				screen.Draw(now)
			}

		case <-quotesQueue.C:
			if !showingHelp && !paused && len(keyboardQueue) == 0 {
				//res := stock.G_STOCK_MANAGER.StockList
				//fmt.Println(res)
				//if res["sh600000"].Market.Open != 0 {
				//	fmt.Println("got quotes")
				go quotes.Fetch()
				redrawQuotesFlag = true
				//}
			}

		case <-marketQueue.C:
			if !showingHelp && !paused {
				screen.Draw(market)
			}

		case msg := <-messages:
			//fmt.Printf("Received message: %s\n", msg.Payload())
			screen.Close()
			// 清除屏幕
			
			if strings.HasPrefix(msg.Topic(), "stock/image/"){
				if strings.HasPrefix(msg.Topic(), "stock/image/day"){
					fmt.Print("\033[2J") 
					os.Stdout.Write(msg.Payload())
					
					break
				}else if strings.HasPrefix(msg.Topic(), "stock/image/time"){
					os.Stdout.Write(msg.Payload())

					preset := &Preset{}
					jsonstr := getuserinput(preset, canstock)
					if jsonstr != "" {
						token := client.Publish("stock/request/"+account, 0, false, jsonstr)
						token.Wait()
					}
				}
			}else if strings.HasPrefix(msg.Topic(), "stock/response"){
				fmt.Print("\033[2J") 
				showposition(string(msg.Payload()) ,si)
				
				preset := &Preset{}
				jsonstr := getuserinput(preset, canstock)
				if jsonstr != "" {
					token := client.Publish("stock/request/"+account, 0, false, jsonstr)
					token.Wait()
				}
			}

			//time.Sleep(1 * time.Second)
			screen := mop.NewScreen(profile)
			defer screen.Close()
		}

		if redrawQuotesFlag && len(keyboardQueue) == 0 {
			screen.DrawOldQuotes(quotes)
			linetxt := screen.Getline3()

			selstock := quotes.GetselectedinfobyTicker(linetxt)
			if selstock != nil {
				canstock = *selstock
			}else{
				canstock = mop.Stock{}
			}
			screen.Setline3(canstock.Ticker)
			redrawQuotesFlag = false
		}
		if redrawMarketFlag && len(keyboardQueue) == 0 {
			screen.Draw(market)
			redrawMarketFlag = false
		}		
	}
}

type Position struct {
    Scode       string  `json:"scode"`
    Sname       string  `json:"sname"`
    Openprice   float64 `json:"openprice"`
    Floatprofit float64 `json:"floatprofit"`
    Marketvalue float64 `json:"marketvalue"`
}

type Preset struct {
	Direction 	int64  `json:"direction"`
    Scode       string  `json:"scode"`
    Sname       string  `json:"sname"`
    Ifabove 	float64 `json:"ifabove"`
    Ifbelow 	float64 `json:"ifbelow"`
	Vol 		float64	`json:"vol"`
}

type Positions struct {
    Position []Position `json:"position"`
	Preset []Preset `json:"preset"`
	Latestinfo string `json:"lastinfo"`
}

func float2Str(f float64) string {
    return fmt.Sprintf("%.2f", f)
}

func showposition(payload string,si map[string]*stock.Stock) string{

	var positions Positions
	//fmt.Println(payload)
    if err := json.Unmarshal([]byte(payload), &positions); err != nil {
        fmt.Println("Error parsing JSON:", err)
        return ""
    }

    var data [][]string
	var totalFloatProfit float64 = 0.0
	var totalMarketValue float64 = 0.0
	//fmt.Println(positions.Position)
    for _, pos := range positions.Position {
        row := []string{pos.Scode + " " +pos.Sname, fmt.Sprintf("%.2f", pos.Openprice), fmt.Sprintf("%.2f", pos.Floatprofit), fmt.Sprintf("%.2f", pos.Marketvalue)}
        data = append(data, row)
		totalFloatProfit += pos.Floatprofit
		totalMarketValue += pos.Marketvalue
    }

	for _, pos := range positions.Preset {
		condition := ""
		newscode := pos.Scode
		row := []string{}
		if pos.Ifabove != 0 {
			condition = "> " + fmt.Sprintf("%.2f", pos.Ifabove)
		}else{
			condition = "< " + fmt.Sprintf("%.2f", pos.Ifbelow)
		}
		//if pos.Scode starts with 's'
		if !strings.HasPrefix(pos.Scode, "s"){
			if strings.HasPrefix(pos.Scode, "6"){
				newscode = "sh" + pos.Scode
			}else{
				newscode = "sz" + pos.Scode
			}
		}
		if pos.Direction == 23 {
        	row = []string{pos.Scode+" "+si[newscode].Base.Name , "buy", condition, fmt.Sprintf("%.2f", pos.Vol)}
		}else{
			row = []string{pos.Scode+" "+si[newscode].Base.Name , "sell", condition, fmt.Sprintf("%.2f", pos.Vol)}
		}
        data = append(data, row)
    }
	
	tableString := &strings.Builder{}
    table := tablewriter.NewWriter(tableString)
	table.SetHeader([]string{"Instrument", "Avg. Price", "Float Profit", "Market Value"})
	table.SetFooter([]string{"", "Total", float2Str(totalFloatProfit), float2Str(totalMarketValue)}) // Add Footer
	//table.EnableBorder(false)                                // Set Border to false
	
	table.SetHeaderColor(tablewriter.Colors{tablewriter.Bold, tablewriter.BgGreenColor},
		tablewriter.Colors{tablewriter.FgHiRedColor, tablewriter.Bold, tablewriter.BgBlackColor},
		tablewriter.Colors{tablewriter.BgRedColor, tablewriter.FgWhiteColor},
		tablewriter.Colors{tablewriter.BgCyanColor, tablewriter.FgWhiteColor})
	
	table.SetColumnColor(tablewriter.Colors{tablewriter.Bold, tablewriter.FgHiBlackColor},
		tablewriter.Colors{tablewriter.Bold, tablewriter.FgHiRedColor},
		tablewriter.Colors{tablewriter.Bold, tablewriter.FgHiBlackColor},
		tablewriter.Colors{tablewriter.Bold, tablewriter.FgBlackColor})
	
	table.SetFooterColor(tablewriter.Colors{}, tablewriter.Colors{},
		tablewriter.Colors{tablewriter.Bold},
		tablewriter.Colors{tablewriter.FgHiRedColor})
	
	table.AppendBulk(data)
	table.Render()
	fmt.Println(tableString.String())//()
	fmt.Println(positions.Latestinfo)
	return tableString.String()
}
//-----------------------------------------------------------------------------
func main() {
	usr, err := user.Current()
	if err != nil {
		panic(err)
	}
	var profileName string
	flag.StringVar(&profileName, "profile", path.Join(usr.HomeDir, defaultProfile), "path to profile")
	var mode string
	flag.StringVar(&mode, "mode", "normal", "execution mode (normal/review)")
	flag.Parse()

	profile, err := mop.NewProfile(profileName)
	if err != nil {
		fmt.Fprintf(os.Stderr, "The profile read from `%s` is corrupted.\n\tError: %s\n\n", profileName, err)

		// Loop until we get a "y" or "n" answer.
		// Note: This is only for the interactive mode. Once we have the "one-shot", this should be skipped
		for {
			fmt.Fprintln(os.Stderr, "Do you want to overwrite the current profile with the default one? [y/n]")
			rne, _, _ := keyboard.GetSingleKey()
			res := strings.ToLower(string(rne))
			if res != "y" && res != "n" {
				fmt.Fprintf(os.Stderr, "Invalid answer `%s`\n\n", res)
				continue
			}

			if res == "y" {
				profile.InitDefaultProfile()
				break
			} else {
				os.Exit(1)
			}
		}
	}

	profile.SetMode(mode)
	fmt.Println("mode:", mode)
	
	if mode == "review" {
		scanner := bufio.NewScanner(os.Stdin)

		// Prompt the user to enter the first date
		fmt.Print("Enter the first date (YYYY-MM-DD): ")
		scanner.Scan()
		firstDateStr := scanner.Text()
	
		// Parse the first date
		firstDate, err := time.Parse("2006-01-02", firstDateStr)
		if err != nil {
			fmt.Println("Error parsing start date:", err)
			return
		}
	
		// Prompt the user to enter the second date
		fmt.Print("Enter the end date (YYYY-MM-DD): ")
		scanner.Scan()
		secondDateStr := scanner.Text()
	
		// Parse the second date
		secondDate, err := time.Parse("2006-01-02", secondDateStr)
		if err != nil {
			fmt.Println("Error parsing end date:", err)
			return
		}
	
		// Print the two dates
		fmt.Println("Start date:", firstDate)
		fmt.Println("End date:", secondDate)
		for d := firstDate; d.Before(secondDate); d = d.AddDate(0, 0, 1) {
			profile.AddDate(d.Format("2006-01-02"))
		}
		//fmt.Println("Dates:", profile.date_json)
	}

	screen := mop.NewScreen(profile)
	defer screen.Close()

	mainLoop(screen, profile, mode)
	profile.Save()
}
